# -*- coding: utf-8 -*-

"""Script for service routines connected with Luci.

Actions available:
  * set initial administrator's password/change current one
  * import/export of the local database

"""

__version__ = '0.1'


#
# GLOBALS
#

LUCI_INI_FILE='/var/lib/luci/etc/luci.ini'


#
# EXCEPTIONS
#

class LuciUtilError(Exception):
    """Just for creating a 'identifiable type' of an exception."""
    pass


#
# HELPER FUNCTIONS
#

def _initOptionParser():
    """Create an OptionParser instance and initialize it.

    Apart from it, it also serves to collect mutually exclusive options
    to the lists of them, so this situation can be easily discovered
    afterwards.

    Return values:
        (parser, mutex_opts_list) pair, where:
            parser    Passed instance of parser with some command-line options
                      added.
            mutex_opts_list
                      List of lists of mutually exclusive options.

    """
    from optparse import OptionParser, OptionGroup

    USAGE = '%prog [-c <config file>] [-i <path>|-e <path>|-p <password>]\n' \
            'Use -h or --help to display details.'
    VERSION = '%%prog %s' % __version__
    parser = OptionParser(usage=USAGE, version=VERSION)

    mutex_opts_list = []
    mutex_opts_list.append([])
    mutex_opts_index = 0

    # ---
    group_db = OptionGroup(parser, 'Whole database related options')
    group_db.add_option('-i', '--import-db',
                        help='import existing database from file',
                        metavar='<path to DB file to import>',
                        action='store', type='string', dest='import_src')
    group_db.add_option('-e', '--export-db',
                        help='export current database to file',
                        metavar='<path to DB file where to export>',
                        action='store', type='string', dest='export_dest')
    mutex_opts_list[mutex_opts_index].append(group_db.get_option('-i'))
    mutex_opts_list[mutex_opts_index].append(group_db.get_option('-e'))
    # ---
    group_misc = OptionGroup(parser, 'Miscellaneous options')
    group_misc.add_option('-p', '--change-password',
                          help='change password (use \'-\' for direct input' \
                               ' afterwards)',
                          metavar='<password>',
                          action='store', type='string', dest='pwd')
    mutex_opts_list[mutex_opts_index].append(group_db.get_option('-p'))
    # ---
    parser.add_option_group(group_db)
    parser.add_option_group(group_misc)

    parser.add_option('-c', '--config',
                      help='Default is %s' % LUCI_INI_FILE,
                      metavar='<config file for Luci>',
                      action='store', type='string', dest='config_file')

    return parser, mutex_opts_list


def _getTG2ConfigParser(config_file):
    """Create and return (Safe)ConfigParser instance with config file loaded.

    If given path to config file is not an absolute path, its supposed that
    given path is relative to DEFAULT_INI_LOCATION.

    Keyword arguments:
        config_file    Path to the config file (usually with .ini extension).

    Return values:
        Created (Safe)ConfigParser instance with config file loaded.

    Raises:
        LuciUtilError

    """
    from os.path import isabs, split, normpath, join, abspath
    from ConfigParser import SafeConfigParser

    if not isabs(config_file):
        config_file = join(abspath(DEFAULT_INI_LOCATION), config_file)

    config_file = normpath(config_file)
    config_file_dir = split(config_file)[0]

    parser = SafeConfigParser({'here': config_file_dir})

    if len(parser.read(config_file)) == 0:
        raise LuciUtilError, 'Could not find configuration file %s.' \
                             % config_file

    return parser


def _SQLAlchemyURLFromConfig(config_file):
    """Open given config file and get 'sqlalchemy.url' value from [app:main].

    Keyword arguments:
        config_file    Path to the config file (usually with .ini extension).

    Return values:
        'sqlalchemy.url' value obtained from the given config file.

    Raises:
        LuciUtilError

    """
    from ConfigParser import Error
    from os.path import normpath

    try:
        config_parser = _getTG2ConfigParser(config_file)
        sql_alchemy_url = config_parser.get('app:main', 'sqlalchemy.url')
    except Error, e:
        raise LuciUtilError, e

    if not sql_alchemy_url.startswith('sqlite:///'):
        raise LuciUtilError, 'DB URL \`%s\' not supported.' \
                             % sql_alchemy_url

    # More multiplatform -- support for different path separators.
    # Length of 'sqlite:///' is 10.
    sql_alchemy_url = sql_alchemy_url[:10] + normpath(sql_alchemy_url[10:])
    return sql_alchemy_url


def _getDBFileFromSQLAlchemyURL(sql_alchemy_url):
    """Convert 'sqlalchemy.url' value to path where such DB is stored.

    Keyword arguments:
        sql_alchemy_url    Value as given by _SQLAlchemyURLFromConfig().

    Return values:
        Path where the DB defined by 'sqlalchemy.url' is stored.

    Raises:
        LuciUtilError

    """
    sql_alchemy_url = sql_alchemy_url.strip()
    if sql_alchemy_url.startswith('sqlite:///'):
        return sql_alchemy_url[10:]  # Length of 'sqlite:///' is 10.
    else:
        raise LuciUtilError, 'DB URL \`%s\' not supported.' \
                             % sql_alchemy_url


#
# FUNCTIONS FOR PARTICULAR OPTIONS
#

def changePassword(config_file, password=None):
    """Change password for the administrator in the database.

    Keyword arguments:
        config_file   Path to the config file (usually with .ini extension).
        password      The password to be set or None to ask at a run-time.

    Return values:


    Raises:


    """

    # TODO -- Needs to be finished (Jan Pokorny will probably not be engaged
    #                               in that).

    from sqlalchemy import create_engine
    from sqlalchemy import Table, Column, Integer, String, MetaData
    from sqlalchemy.orm import mapper
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm.exc import NoResultFound
    from sqlalchemy.orm import sessionmaker

    from getpass import getpass, GetPassWarning
    from luci.model.auth import User, Group, Permission

    try:
        sqlalchemy_url = _SQLAlchemyURLFromConfig(config_file)
    except LuciUtilError, e:
        print e
    else:
        print sqlalchemy_url
        engine = create_engine(sqlalchemy_url)
        session = sessionmaker(bind=engine)
        metadata = MetaData()
        metadata.create_all(engine)

        # Try to get the current data for given user. If it raises error,
        # such column/table has to be created first.
        try:
            result = session.query(User).filter_by(user_name=u'admin').one()
        except NoResultFound, e:
            print e


def importDatabase(config_file, source, backup_old=False):
    """Import the database from given source (replace the old one if present).

    If requested, the old DB file can be backed up.

    Keyword arguments:
        config_file   Path to the config file (usually with .ini extension).
        source        Path of source DB file to be imported.
        backup_old    Whether to backup the old DB if was in place.

    Raises:
        LuciUtilError
        ...

    """
    from os import rename
    from os.path import exists
    from datetime import datetime
    from shutil import copy

    SQLAlchemy_url = _SQLAlchemyURLFromConfig(config_file)
    DB_path = _getDBFileFromSQLAlchemyURL(SQLAlchemy_url)

    if not exists(source):
        raise LuciUtilError, 'Source file for import does not exists: %s' \
                             % source

    if backup_old and exists(DB_path):
        rename(DB_path, DB_path + '.' + datetime.now().strftime('%y%m%d-%H%M'))

    copy(source, DB_path)


def exportDatabase(config_file, destination, overwrite=True):
    """Export the database to given destination.

    If requested, the file with the same name as a destination file
    is NOT overwritten (if exists).

    Keyword arguments:
        config_file  Path to the config file (usually with .ini extension).
        destination  Path to destination DB file, where to export DB file.
        overwrite    Whether to backup the old DB if was in place.

    Raises:
        LuciUtilError
        ...

    """
    from os import rename
    from os.path import exists
    from datetime import datetime
    from shutil import copy

    SQLAlchemy_url = _SQLAlchemyURLFromConfig(config_file)
    DB_path = _getDBFileFromSQLAlchemyURL(SQLAlchemy_url)

    if not exists(DB_path):
        raise LuciUtilError, 'Default DB file does not exist: %s' \
                             % DB_path

    if not exists(destination) or overwrite:
        copy(DB_path, destination)


#
# MAIN
#

def main():
    """Main body of the script."""

    (parser, mutex_opts_list) = _initOptionParser()
    (options, args) = parser.parse_args()

    # No stand-alone argument (i.e. without preceding option) allowed.
    if len(args) != 0:
        parser.print_usage()
        exit(1)

    # Check, whether mutual exclusivity of options is kept.
    for mutex_opts in mutex_opts_list:
        i = 0
        for mutex_opt in mutex_opts:
            if getattr(options, mutex_opt.dest, False):
                i += 1
        if i > 1:
            mutex_opts_str = ['\'' + opt._short_opts[0] + '\'' \
                              for opt in mutex_opts]
            parser.error('all options %s are mutually exclusive' % \
                         ', '.join(mutex_opts_str))

    # Set the configuration file to be used according to whether it is defined
    # explicitly from command-line options or not.
    if options.config_file:
        config_file = options.config_file
    else:
        config_file = LUCI_INI_FILE

    # Branch according to the option used.
    if options.pwd:
        # Change password.
        try:
            # Minus as a password means to ask for it afterwards.
            if options.pwd == '-':
                changePassword(config_file)
            else:
                changePassword(config_file, options.pwd)
        except LuciUtilError, e:
            print e
    elif options.import_src:
        # Import database.
        try:
            importDatabase(config_file, options.import_src) # backup_old=True
        except LuciUtilError, e:
            print e
    elif options.export_dest:
        # Export database.
        try:
            exportDatabase(config_file, options.export_dest)
        except LuciUtilError, e:
            print e
    else:
        # No option used.
        parser.print_usage()
        exit(1)

# -----------------------------------------------------------------------------

if __name__ == '__main__':
    # Module executed as a script.
    main()
